import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:hydrated_bloc/hydrated_bloc.dart';

import 'package:puissance4/models/game/game.dart';
import 'package:puissance4/models/settings/settings_game.dart';
import 'package:puissance4/models/settings/settings_global.dart';
import 'package:puissance4/robot/robot_player.dart';
import 'package:puissance4/utils/tools.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createNull(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      isFinished: state.currentGame.isFinished,
      animationInProgress: state.currentGame.animationInProgress,
      // Base data
      board: state.currentGame.board,
      sizeHorizontal: state.currentGame.sizeHorizontal,
      sizeVertical: state.currentGame.sizeVertical,
      // Game data
      currentPlayer: state.currentGame.currentPlayer,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      // Settings
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    refresh();

    robotPlay();
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    state.currentGame.isFinished = true;
    refresh();
  }

  void toggleCurrentPlayer() {
    state.currentGame.currentPlayer = 3 - state.currentGame.currentPlayer;
    refresh();

    robotPlay();
  }

  void robotPlay() async {
    if (!state.currentGame.isFinished && !state.currentGame.isCurrentPlayerHuman()) {
      final int? pickedCell = RobotPlayer.pickColumn(state.currentGame);
      await Future.delayed(const Duration(milliseconds: 500));

      if (pickedCell != null) {
        tapOnColumn(col: pickedCell);
      }
    }
  }

  void tapOnColumn({required int col}) async {
    printlog('tap on column: $col');

    if (!state.currentGame.isMoveAllowed(col: col)) {
      printlog('not allowed');
      return;
    }

    state.currentGame.animationInProgress = true;
    refresh();

    await dropCoin(col: col);

    if (state.currentGame.connected4()) {
      printlog('user connected 4');
      state.currentGame.isFinished = true;
      refresh();
    }
    if (!state.currentGame.canPlay()) {
      printlog('user has no more move to play');
      state.currentGame.isFinished = true;
      refresh();
    }

    state.currentGame.animationInProgress = false;
    if (state.currentGame.isFinished == false) {
      toggleCurrentPlayer();
    }

    refresh();
  }

  // put coin on top cell
  // and make it fall until bottom or other coin
  Future<void> dropCoin({required int col}) async {
    int currentRow = 0;

    do {
      if (currentRow > 0) {
        state.currentGame.board.setCellValue(
          row: currentRow - 1,
          col: col,
          value: 0,
        );
      }
      state.currentGame.board.setCellValue(
        row: currentRow++,
        col: col,
        value: state.currentGame.currentPlayer,
      );
      refresh();

      await Future.delayed(const Duration(milliseconds: 50));
    } while (currentRow < state.currentGame.sizeVertical &&
        state.currentGame.board.getCellValue(
              row: currentRow,
              col: col,
            ) ==
            0);

    refresh();
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame.toJson(),
    };
  }
}
