import 'package:puissance4/config/default_game_settings.dart';
import 'package:puissance4/models/game/board.dart';
import 'package:puissance4/models/settings/settings_game.dart';
import 'package:puissance4/models/settings/settings_global.dart';
import 'package:puissance4/utils/tools.dart';

class Game {
  Game({
    // Settings
    required this.gameSettings,
    required this.globalSettings,

    // State
    this.isRunning = false,
    this.isStarted = false,
    this.isFinished = false,
    this.animationInProgress = false,

    // Base data
    required this.board,
    required this.sizeHorizontal,
    required this.sizeVertical,

    // Game data
    required this.currentPlayer,
  });

  // Settings
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  // State
  bool isRunning;
  bool isStarted;
  bool isFinished;
  bool animationInProgress;

  // Base data
  Board board;
  final int sizeHorizontal;
  final int sizeVertical;

  // Game data
  int currentPlayer;

  factory Game.createNull() {
    return Game(
      // Settings
      gameSettings: GameSettings.createDefault(),
      globalSettings: GlobalSettings.createDefault(),
      // Base data
      board: Board.createNull(),
      sizeHorizontal: 0,
      sizeVertical: 0,
      // Game data
      currentPlayer: 0,
    );
  }

  factory Game.createNew({
    GameSettings? gameSettings,
    GlobalSettings? globalSettings,
  }) {
    final GameSettings newGameSettings = gameSettings ?? GameSettings.createDefault();
    final GlobalSettings newGlobalSettings = globalSettings ?? GlobalSettings.createDefault();

    final int sizeHorizontal = int.parse(newGameSettings.size.split('x')[0]);
    final int sizeVertical = int.parse(newGameSettings.size.split('x')[1]);

    // Create empty board (it will be mined after first hit)
    final BoardCells board = [];
    for (int rowIndex = 0; rowIndex < sizeVertical; rowIndex++) {
      final List<int> row = [];
      for (int colIndex = 0; colIndex < sizeHorizontal; colIndex++) {
        row.add(0);
      }
      board.add(row);
    }

    return Game(
      // Settings
      gameSettings: newGameSettings,
      globalSettings: newGlobalSettings,
      // State
      isRunning: true,
      // Base data
      board: Board.createNew(cells: board),
      sizeHorizontal: sizeHorizontal,
      sizeVertical: sizeVertical,
      // Game data
      currentPlayer: 1,
    );
  }

  bool get canBeResumed => isStarted && !isFinished;

  // Ensure move is allowed
  bool isMoveAllowed({required int col}) {
    // ensure first row is empty on selected col
    return board.getCellValue(row: 0, col: col) == 0;
  }

  bool connected4() {
    // vertical
    for (int row = 0; row < sizeVertical - 3; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if ((board.getCellValue(row: row, col: col) == currentPlayer) &&
            (board.getCellValue(row: row + 1, col: col) == currentPlayer) &&
            (board.getCellValue(row: row + 2, col: col) == currentPlayer) &&
            (board.getCellValue(row: row + 3, col: col) == currentPlayer)) {
          return true;
        }
      }
    }

    // horizontal
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal - 3; col++) {
        if ((board.getCellValue(row: row, col: col) == currentPlayer) &&
            (board.getCellValue(row: row, col: col + 1) == currentPlayer) &&
            (board.getCellValue(row: row, col: col + 2) == currentPlayer) &&
            (board.getCellValue(row: row, col: col + 3) == currentPlayer)) {
          return true;
        }
      }
    }

    // diagonal down
    for (int row = 0; row < sizeVertical - 3; row++) {
      for (int col = 0; col < sizeHorizontal - 3; col++) {
        if ((board.getCellValue(row: row, col: col) == currentPlayer) &&
            (board.getCellValue(row: row + 1, col: col + 1) == currentPlayer) &&
            (board.getCellValue(row: row + 2, col: col + 2) == currentPlayer) &&
            (board.getCellValue(row: row + 3, col: col + 3) == currentPlayer)) {
          return true;
        }
      }
    }

    // diagonal up
    for (int row = 0; row < sizeVertical - 3; row++) {
      for (int col = 0; col < sizeHorizontal - 3; col++) {
        if ((board.getCellValue(row: row + 3, col: col) == currentPlayer) &&
            (board.getCellValue(row: row + 2, col: col + 1) == currentPlayer) &&
            (board.getCellValue(row: row + 1, col: col + 2) == currentPlayer) &&
            (board.getCellValue(row: row, col: col + 3) == currentPlayer)) {
          return true;
        }
      }
    }

    return false;
  }

  bool canPlay() {
    List<int> allowedColumns = [];

    for (int columnIndex = 0; columnIndex < sizeHorizontal; columnIndex++) {
      if (isMoveAllowed(col: columnIndex)) {
        allowedColumns.add(columnIndex);
      }
    }

    return allowedColumns.isNotEmpty;
  }

  bool isPlayerHuman(int playerIndex) {
    switch (gameSettings.gameMode) {
      case DefaultGameSettings.gameModeHumanVsHuman:
        return true;
      case DefaultGameSettings.gameModeHumanVsRobot:
        return (playerIndex == 1);
      case DefaultGameSettings.gameModeRobotVsHuman:
        return (playerIndex == 2);
      case DefaultGameSettings.gameModeRobotVsRobot:
        return false;
      default:
    }
    return true;
  }

  bool isCurrentPlayerHuman() {
    return isPlayerHuman(currentPlayer);
  }

  void dump() {
    printlog('');
    printlog('## Current game dump:');
    printlog('');
    printlog('$Game:');
    printlog('  Settings');
    gameSettings.dump();
    globalSettings.dump();
    printlog('  State');
    printlog('    isRunning: $isRunning');
    printlog('    isStarted: $isStarted');
    printlog('    isFinished: $isFinished');
    printlog('    animationInProgress: $animationInProgress');
    board.dump();
    printlog('  Game data');
    printlog('    currentPlayer: $currentPlayer');
    printlog('');
  }

  @override
  String toString() {
    return '$Game(${toJson()})';
  }

  Map<String, dynamic>? toJson() {
    return <String, dynamic>{
      // Settings
      'gameSettings': gameSettings.toJson(),
      'globalSettings': globalSettings.toJson(),
      // State
      'isRunning': isRunning,
      'isStarted': isStarted,
      'isFinished': isFinished,
      'animationInProgress': animationInProgress,
      // Base data
      'board': board.toJson(),
      // Game data
      'currentPlayer': currentPlayer,
    };
  }
}
